"""
Contains functions that are shared amongst the children of the HappyTrainer class.

Based on
https://github.com/huggingface/transformers/blob/master/examples/pytorch/language-modeling/run_mlm.py
and
https://github.com/huggingface/transformers/blob/master/examples/pytorch/language-modeling/run_clm.py
"""


def preprocess_concatenate(tokenizer, dataset, preprocessing_processes, mlm=True):
    """
    :param tokenizer: tokenizer for a transformer model
    :param dataset: a datasets.Dataset object
    :param preprocessing_processes:
    :param mlm:

    :return:
    """

    max_input_length = tokenizer.model_max_length


    def tokenize_function(example):
        return tokenizer(example["text"])

    tokenized_dataset = dataset.map(tokenize_function, batched=True,
                                      num_proc=preprocessing_processes,
                                      remove_columns=["text"])

    def group_texts(examples):
        concatenated_examples = {k: sum(examples[k], []) for k in examples.keys()}
        total_length = len(concatenated_examples[list(examples.keys())[0]])
        output_length = (total_length // max_input_length) * max_input_length

        # if  total_length is less than the max_input_length length
        # then it causes an error unless the code below is used.
        # this is due to total_length being truncated to 0
        if output_length == 0:
            output_length = total_length

        result = {
            k: [t[i: i + max_input_length] for i in range(0, output_length, max_input_length)]
            for k, t in concatenated_examples.items()
        }

        if not mlm:
            # Masked language models don't need labels. Text generation models do
            result["labels"] = result["input_ids"].copy()
        return result

    tokenized_dataset = tokenized_dataset.map(
        group_texts,
        batched=True,
        num_proc=preprocessing_processes,
    )


    return tokenized_dataset



def create_args_dataclass(default_dic_args, input_dic_args, method_dataclass_args):
    """
    Combines default_dic_args and input_dic_args and then outputs a dataclass.

    The values of input_dic_args overwrite the values of default_dic_args.

    default_dic_args and dataclass_args must have the same keys.

    :param default_dic_args: A dictionary that contains default settings. Example: ARGS_WP_EVAl
    :param input_dic_args: A dictionary a user inputs for **kwargs when using .train(), .eval() or .test()
    :param method_dataclass_args: A class for target functionality. Example: WPEvalArgs
    :return: A dataclass object that will then be passed to HappyTrainer.train()/eval/test
    """
    settings_dic = {**default_dic_args, **input_dic_args}
    return method_dataclass_args(**settings_dic)


"""
Contains the parent class to HappyTextClassification, HappyWordPrediction, HappyQuestionAnswering
and HappyNextSentencePrediction called HappyTransformer

Contains shared variables and methods for these classes.
"""
import logging
import torch
from transformers import  AutoTokenizer, AutoConfig

class HappyTransformer():
    """
    Parent class to HappyTextClassification, HappyWordPrediction, HappyQuestionAnswering
    and HappyNextSentencePrediction.

    """

    def __init__(self, model_type, model_name, model, load_path=""):
        self.model_type = model_type  # BERT, #DISTILBERT, ROBERTA, ALBERT etc
        self.model_name = model_name

        if load_path != "":
            self.tokenizer = AutoTokenizer.from_pretrained(load_path)
        else:
            self.tokenizer = AutoTokenizer.from_pretrained(model_name)
        self.model = model
        self.model.eval()
        self._trainer = None  # initialized in child class

        # todo  change logging system
        self.logger = logging.getLogger(__name__)

        handler = logging.StreamHandler()
        handler.addFilter(logging.Filter('happytransformer'))
        logging.basicConfig(
            format='%(asctime)s - %(levelname)s - %(name)s -   %(message)s',
            datefmt='%m/%d/%Y %H:%M:%S',
            level=logging.INFO,
            handlers=[handler]
        )

        self._device = torch.device(
            "cuda" if torch.cuda.is_available()
            else "cpu"
        )

        if self._device == 'cuda':
            self.model.to(self._device)
        self.logger.info("Using model: %s", self._device)


    def train(self, input_filepath, args):
        """
        Trains a model
        :param input_filepath: a string that contains a path to a csv file
         that contains testing data
        :param args: settings in the form of a dictionary
        :return: None
        """
        raise NotImplementedError()

    def eval(self, input_filepath, args):
        """
        Evaluates the model. Determines how well the model performs on a given dataset
        :param input_filepath: a string that contains a path to a
         csv file that contains evaluating data
        :param args: settings in the form of a dictionary
        :return: correct percentage
        """
        raise NotImplementedError()

    def test(self, input_filepath, args):
        """
        Used to generate predictions for a given dataset.
        The dataset may not be labelled.
        :param args: settings in the form of a dictionary

        :param input_filepath: a string that contains a path to
        a csv file that contains testing data

        """
        raise NotImplementedError()

    def save(self, path):
        """
        Saves both the model, tokenizer and various configuration/settings files
        to a given path

        :param path: string:  a path to a directory
        :return:
        """
        self.model.save_pretrained(path)
        self.tokenizer.save_pretrained(path)




def run_save_load(happy, output_path, args, data_path, run_type):
    args['save_preprocessed_data_path'] = output_path
    args['load_preprocessed_data_path'] = output_path

    args['save_preprocessed_data'] = True
    if run_type == "train":
        happy.train(data_path, args=args)
    elif run_type == "eval":
        result = happy.eval(data_path, args=args)
    elif run_type == "test":
        result = happy.test(data_path, args=args)
    else:
        ValueError("invalid run_type for run_save_load")

    args['save_preprocessed_data'] = False
    args['load_preprocessed_data'] = True
    if run_type == "train":
        happy.train(data_path, args=args)
    elif run_type == "eval":
        result = happy.eval(data_path, args=args)
    elif run_type == "test":
        result = happy.test(data_path, args=args)





"""
Parent class for training classes, such as TCTrainer and QATrainer
"""
from dataclasses import dataclass
import tempfile
from transformers import TrainingArguments, Trainer

@dataclass
class EvalResult:
    loss: float

class HappyTrainer:
    def __init__(self, model, model_type, tokenizer, device, logger):
        self.model = model
        self.model_type = model_type
        self.tokenizer = tokenizer
        self.device = device
        self.logger = logger

    def train(self, input_filepath, args):
        """

        :param input_filepath: A string to file location
        :param args: a dictionary that contains settings
        :return:
        """
        raise NotImplementedError()

    def test(self, input_filepath, solve, args):
        """

        :param input_filepath: A string to file location
        :param solve: a method for using the model for the given task
        :return: test results
        """
        raise NotImplementedError()

    def eval(self, input_filepath, args):
        """
        :param input_filepath: A string to file location
        :args a dictionary that contains settings
        :return: a dictionary that contains a key called "eval_loss" that holds the loss
         for the given eval dataset. May add more metrics later
        """
        raise NotImplementedError()

    @staticmethod
    def _get_data(filepath, test_data=False):
        """

        :param filepath:  A string to file location
        :param test_data: False for train and eval, True for test
        :return: varies for each task
        """
        raise NotImplementedError()

    @staticmethod
    def _get_training_args(dataclass_args, output_path):
        """
        :param args: a dataclass of arguments for training
        :param output_path: A string to a temporary directory
        :return: A TrainingArguments object
        """
        return TrainingArguments(
            output_dir=output_path,
            learning_rate=dataclass_args.learning_rate,
            weight_decay=dataclass_args.weight_decay,
            adam_beta1=dataclass_args.adam_beta1,
            adam_beta2=dataclass_args.adam_beta2,
            adam_epsilon=dataclass_args.adam_epsilon,
            max_grad_norm=dataclass_args.max_grad_norm,
            num_train_epochs=dataclass_args.num_train_epochs,
            report_to=["none"],
            per_device_train_batch_size=dataclass_args.batch_size

        )


    def _run_train(self, dataset, dataclass_args, data_collator):
        """

        :param dataset: a child of torch.utils.data.Dataset
        :param dataclass_args: a dataclass that contains settings
        :return: None
        """
        with tempfile.TemporaryDirectory() as tmp_dir_name:
            training_args = self._get_training_args(dataclass_args, tmp_dir_name)
            trainer = Trainer(
                model=self.model,
                args=training_args,
                train_dataset=dataset,
                tokenizer=self.tokenizer,
                data_collator=data_collator,
            )
            trainer.train()
    def _run_eval(self, dataset, data_collator, dataclass_args):
        """
        :param dataset: a child of torch.utils.data.Dataset
        :return: None
        """
        with tempfile.TemporaryDirectory() as tmp_dir_name:
            eval_args = self._get_eval_args(tmp_dir_name, dataclass_args)
            trainer = Trainer(
                model=self.model,
                args=eval_args,
                eval_dataset=dataset,
                data_collator=data_collator
            )
            return trainer.evaluate()

    @staticmethod
    def _get_eval_args(output_path, dataclass_args):
        """
        :param output_path: A string to a temporary directory
        :return: A TrainingArguments object
        """
        return TrainingArguments(
            output_dir=output_path,
            seed=42,
            report_to=['none'],
            per_device_eval_batch_size=dataclass_args.batch_size,

        )



ARGS_QA_TRAIN = {
    'learning_rate': 5e-5,
    'num_train_epochs': 3.0,
    'batch_size': 1,
    'weight_decay': 0,
    'adam_beta1': 0.9,
    'adam_beta2': 0.999,
    'adam_epsilon': 1e-8,
    'max_grad_norm':  1.0,
    'save_preprocessed_data': False,
    'save_preprocessed_data_path': "",
    'load_preprocessed_data': False,
    'load_preprocessed_data_path': "",
}


ARGS_QA_EVAl = {
    'save_preprocessed_data': False,
    'save_preprocessed_data_path': "",
    'load_preprocessed_data': False,
    'load_preprocessed_data_path': "",
    'batch_size': 1,

}

ARGS_QA_TEST = {
    'save_preprocessed_data': False,
    'save_preprocessed_data_path': "",
    'load_preprocessed_data': False,
    'load_preprocessed_data_path': "",
}


from .adaptor import Adaptor
from .berts import RobertaAdaptor, AlbertAdaptor

ADAPTORS = {
    'ALBERT': AlbertAdaptor(),
    'ROBERTA': RobertaAdaptor(),

}

def get_adaptor(model_type:str)->Adaptor:
    if model_type in ADAPTORS:
        return ADAPTORS[model_type]
    else:
        # Default for models with no special cases
        return Adaptor()



"""
Tests for Text Classification Functionality
"""

from happytransformer import(
    HappyTextClassification,
    TCTrainArgs,
    TCEvalArgs,
    TCTestArgs,
    ARGS_TC_TRAIN,
    ARGS_TC_EVAL
)
from happytransformer.happy_text_classification import TextClassificationResult
from tests.shared_tests import run_save_load
from pytest import approx



def test_classify_text():
    happy_tc = HappyTextClassification(model_type="DISTILBERT", model_name="distilbert-base-uncased-finetuned-sst-2-english")
    result = happy_tc.classify_text("What a great movie")
    assert result.label == 'POSITIVE'
    assert result.score > 0.9


def test_tc_train():
    happy_tc = HappyTextClassification(
        model_type="DISTILBERT",
        model_name="distilbert-base-uncased-finetuned-sst-2-english"
    )
    results = happy_tc.train("../data/tc/train-eval.csv")


def test_tc_eval():
    happy_tc = HappyTextClassification(
        model_type="DISTILBERT",
        model_name="distilbert-base-uncased-finetuned-sst-2-english"
    )
    results = happy_tc.eval("../data/tc/train-eval.csv")
    assert results.loss == approx(0.007262040860950947, 0.01)


def test_tc_test():
    happy_tc = HappyTextClassification(
        model_type="DISTILBERT",
        model_name="distilbert-base-uncased-finetuned-sst-2-english"
    )

    result = happy_tc.test("../data/tc/test.csv")
    answer = [
        TextClassificationResult(label='POSITIVE', score=0.9998401999473572),
        TextClassificationResult(label='NEGATIVE', score=0.9772131443023682),
        TextClassificationResult(label='NEGATIVE', score=0.9966067671775818),
        TextClassificationResult(label='POSITIVE', score=0.9792295098304749)
    ]
    assert result == answer


def test_tc_train_effectiveness():
    """assert that training decreases the loss"""
    happy_tc = HappyTextClassification(
        model_type="DISTILBERT",
        model_name="distilbert-base-uncased"
    )
    before_loss = happy_tc.eval("../data/tc/train-eval.csv").loss
    happy_tc.train("../data/tc/train-eval.csv")
    after_loss = happy_tc.eval("../data/tc/train-eval.csv").loss
    assert after_loss < before_loss


def test_tc_train_effectiveness_multi():
    
    happy_tc = HappyTextClassification(
        model_type="DISTILBERT",
        model_name="distilbert-base-uncased", 
        num_labels=3
    )
    before_loss = happy_tc.eval("../data/tc/train-eval-multi.csv").loss
    happy_tc.train("../data/tc/train-eval-multi.csv")
    after_loss = happy_tc.eval("../data/tc/train-eval-multi.csv").loss
    assert after_loss < before_loss


def test_tc_save():
    happy = HappyTextClassification()
    happy.save("model/")
    result_before = happy.classify_text("What a great movie")

    happy = HappyTextClassification(load_path="model/")
    result_after = happy.classify_text("What a great movie")

    assert result_before.label==result_after.label
    

def test_tc_with_dic():

    happy_tc = HappyTextClassification()
    train_args = {'learning_rate': 0.01,  "num_train_epochs": 1}


    happy_tc.train("../data/tc/train-eval.csv" , args=train_args)

    eval_args = {}

    result_eval = happy_tc.eval("../data/tc/train-eval.csv", args=eval_args)

    test_args = {}

    result_test = happy_tc.test("../data/tc/test.csv", args=test_args)


def test_tc_with_dataclass():

    happy_tc = HappyTextClassification()
    train_args = TCTrainArgs(learning_rate=0.01, num_train_epochs=1)

    happy_tc.train("../data/tc/train-eval.csv", args=train_args)

    eval_args = TCEvalArgs()

    result_eval= happy_tc.eval("../data/tc/train-eval.csv", args=eval_args)


    test_args = TCTestArgs()

    result_test = happy_tc.test("../data/tc/test.csv", args=test_args)

def test_tc_save_load_train():
    happy_wp = HappyTextClassification()
    output_path = "data/tc-train.json"
    data_path = "../data/tc/train-eval.csv"
    run_save_load(happy_wp, output_path, ARGS_TC_TRAIN, data_path, "train")


def test_tc_save_load_eval():
    happy_wp = HappyTextClassification()
    output_path = "data/tc-train.json"
    data_path = "../data/tc/train-eval.csv"
    run_save_load(happy_wp, output_path, ARGS_TC_EVAL, data_path, "eval")



from happytransformer import HappyNextSentence


def test_sp_true():
    happy_ns = HappyNextSentence()
    result = happy_ns.predict_next_sentence(
        "Hi nice to meet you. How old are you?",
        "I am 21 years old."
    )
    assert result > 0.5


def test_sp_false():
    happy_ns = HappyNextSentence()
    result = happy_ns.predict_next_sentence(
        "How old are you?",
        "The Eiffel Tower is in Paris."
    )
    assert result < 0.5

def test_sp_save():
    happy = HappyNextSentence()
    happy.save("model/")
    result_before = happy.predict_next_sentence(
        "How old are you?",
        "The Eiffel Tower is in Paris."
    )

    happy = HappyNextSentence(load_path="model/")
    result_after = happy.predict_next_sentence(
        "How old are you?",
        "The Eiffel Tower is in Paris."
    )

    assert result_before == result_after




"""
Fine-tuning for masked word prediction models.

Based on the tutorial found here:
https://github.com/huggingface/transformers/blob/master/examples/pytorch/language-modeling/run_mlm.py
"""
from dataclasses import dataclass
from transformers import DataCollatorForLanguageModeling
from happytransformer.happy_trainer import HappyTrainer, EvalResult
from datasets import load_dataset
from happytransformer.fine_tuning_util import preprocess_concatenate
from happytransformer.wp.default_args import ARGS_WP_TRAIN, ARGS_WP_EVAl, ARGS_WP_TEST
import json


@dataclass
class WPTrainArgs:
    learning_rate: float = ARGS_WP_TRAIN["learning_rate"]
    num_train_epochs: int = ARGS_WP_TRAIN["num_train_epochs"]
    batch_size: int = ARGS_WP_TRAIN["batch_size"]
    weight_decay: float = ARGS_WP_TRAIN["weight_decay"]
    adam_beta1: float = ARGS_WP_TRAIN["adam_beta1"]
    adam_beta2: float = ARGS_WP_TRAIN["adam_beta2"]
    adam_epsilon: float = ARGS_WP_TRAIN["adam_epsilon"]
    max_grad_norm:  float = ARGS_WP_TRAIN["max_grad_norm"]
    save_preprocessed_data: bool = ARGS_WP_TRAIN["save_preprocessed_data"]
    save_preprocessed_data_path: str = ARGS_WP_TRAIN["save_preprocessed_data_path"]
    load_preprocessed_data: bool = ARGS_WP_TRAIN["load_preprocessed_data"]
    load_preprocessed_data_path: str = ARGS_WP_TRAIN["load_preprocessed_data_path"]
    preprocessing_processes: int = ARGS_WP_TRAIN["preprocessing_processes"]
    mlm_probability: float = ARGS_WP_TRAIN["mlm_probability"]
    line_by_line: bool = ARGS_WP_TRAIN["line_by_line"]


@dataclass
class WPEvalArgs:
    batch_size: int = ARGS_WP_EVAl["batch_size"]
    save_preprocessed_data: bool = ARGS_WP_EVAl["save_preprocessed_data"]
    save_preprocessed_data_path: str = ARGS_WP_EVAl["save_preprocessed_data_path"]
    load_preprocessed_data: bool = ARGS_WP_EVAl["load_preprocessed_data"]
    load_preprocessed_data_path: str = ARGS_WP_EVAl["load_preprocessed_data_path"]
    preprocessing_processes: int =ARGS_WP_EVAl["preprocessing_processes"]
    mlm_probability: float = ARGS_WP_EVAl["mlm_probability"]
    line_by_line: bool = ARGS_WP_EVAl["line_by_line"]



class WPTrainer(HappyTrainer):
    """
    Trainer class for HappyWordPrediction
    """
    def train(self, input_filepath, dataclass_args: WPTrainArgs):
        """
        :param input_filepath: A file path to a text file that contains nothing but training data
        :param dataclass_args: A WPTrainArgs() object
        :return: None
        """
        if not dataclass_args.load_preprocessed_data:
            self.logger.info("Preprocessing dataset...")

            dataset = load_dataset("text", data_files={"train": input_filepath})
            if dataclass_args.line_by_line:
                tokenized_dataset = self._preprocess_line_by_line(self.tokenizer, dataset, dataclass_args.preprocessing_processes)
            else:
                tokenized_dataset = preprocess_concatenate(self.tokenizer, dataset, dataclass_args.preprocessing_processes, True)
        else:
            self.logger.info("Loading dataset from %s...", dataclass_args.load_preprocessed_data_path)
            tokenized_dataset = load_dataset("json", data_files={"train": dataclass_args.load_preprocessed_data_path}, field='train')

        if dataclass_args.save_preprocessed_data:
            if dataclass_args.load_preprocessed_data:
                self.logger.warning("Both save_preprocessed_data and load_data are enabled,")

            self.logger.info("Saving training dataset to %s...", dataclass_args.save_preprocessed_data_path)

            self._generate_json(dataclass_args.save_preprocessed_data_path, tokenized_dataset["train"], "train")

        data_collator = DataCollatorForLanguageModeling(tokenizer=self.tokenizer,
                                                        mlm_probability=dataclass_args.mlm_probability)
        self.logger.info("Training...")

        self._run_train(tokenized_dataset['train'], dataclass_args, data_collator)


    def eval(self, input_filepath, dataclass_args: WPEvalArgs):
        """
        :param input_filepath: A file path to a text file that contains nothing but evaluating data
        :param dataclass_args: A WPEvalArgs() object
        :return: An EvalResult() object
        """
        dataset = load_dataset("text", data_files={"eval": input_filepath})

        if dataclass_args.line_by_line:
            tokenized_dataset = self._preprocess_line_by_line(self.tokenizer, dataset, dataclass_args.preprocessing_processes)
        else:
            tokenized_dataset = preprocess_concatenate(self.tokenizer, dataset, dataclass_args.preprocessing_processes, True)


        data_collator = DataCollatorForLanguageModeling(tokenizer=self.tokenizer,
                                                        mlm_probability=dataclass_args.mlm_probability)
        result = self._run_eval(tokenized_dataset['eval'], data_collator, dataclass_args)

        return EvalResult(loss=result["eval_loss"])


    def test(self, input_filepath, solve, args):
        raise NotImplementedError()


    def _preprocess_line_by_line(self, tokenizer, dataset, preprocessing_processes):
        """
        :param tokenizer: tokenizer for a transformer model
        :param datasets: a datasets.Dataset object
        :param preprocessing_processes: number of processes to use for pre-processing
        :return:
        """

        def tokenize_function(example):
            return tokenizer(example["text"],
                             add_special_tokens=True, truncation=True,)

        tokenized_dataset = dataset.map(tokenize_function, batched=True,
                                          num_proc=preprocessing_processes,
                                          remove_columns=["text"])
        return tokenized_dataset


    def _generate_json(self, json_path, dataset, name):
        """
        :param json_path: A path to a json file that will be created/overwritten
        :param dataset: A list of dictionaries that contain the keys "attention_mask," "input_ids" and "labels"
        :param name: A string to specify if the written data is for "Train" or "Eval"
        :return: None
        """
        data = {}
        data[name] = []
        data = {
            name: [
                {
                    'attention_mask': case['attention_mask'],
                    'input_ids': case['input_ids'],
                }
                for case in dataset
            ]
        }

        with open(json_path, 'w') as outfile:
            json.dump(data, outfile)


"""
Tests for the question answering training, evaluating and testing functionality
"""

from happytransformer.happy_question_answering import (
    HappyQuestionAnswering,
    QATrainArgs,
    QAEvalArgs,
    QATestArgs
)

from pytest import approx

def test_qa_basic():

    happy_qa = HappyQuestionAnswering()
    result = happy_qa.answer_question("Today's date is January 8th 2021", "What is the date?", top_k=10)
    assert result[0].answer == 'January 8th 2021'
    assert result[0].score == approx(0.9696964621543884, 1)

    #assert sum(answer.score for answer in answers) == approx(1, 0.5)
    #assert all('January 8th' in answer.answer for answer in answers)


def test_qa_train():
    happy_qa = HappyQuestionAnswering(
        model_type='DISTILBERT',
        model_name='distilbert-base-cased-distilled-squad'
    )
    result = happy_qa.train("../data/qa/train-eval.csv")


def test_qa_eval():
    happy_qa = HappyQuestionAnswering(
        model_type='DISTILBERT',
        model_name='distilbert-base-cased-distilled-squad'
    )
    result = happy_qa.eval("../data/qa/train-eval.csv")
    assert result.loss == approx(0.11738169193267822, 0.001)


def test_qa_test():
    happy_qa = HappyQuestionAnswering()
    results = happy_qa.test("../data/qa/test.csv")
    assert results[0].answer == 'October 31st'
    assert results[1].answer == 'November 23rd'


def test_qa_train_effectiveness():
    """
    Ensures that HappyQuestionAnswering.train() results in
    lowering the loss as determined by HappyQuestionAnswering.eval()
    """
    # use a non-fine-tuned model so we DEFINITELY get an improvement
    happy_qa = HappyQuestionAnswering()
    before_loss = happy_qa.eval("../data/qa/train-eval.csv").loss
    happy_qa.train("../data/qa/train-eval.csv")
    after_loss = happy_qa.eval("../data/qa/train-eval.csv").loss

    assert after_loss < before_loss


def test_qa_save():
    happy = HappyQuestionAnswering()
    happy.save("model/")
    result_before = happy.answer_question("Natural language processing is a subfield of artificial surrounding creating models that understand language","What is natural language processing?")

    happy = HappyQuestionAnswering(load_path="model/")
    result_after = happy.answer_question("Natural language processing is a subfield of artificial surrounding creating models that understand language","What is natural language processing?")

    assert result_before[0].answer == result_after[0].answer


def test_qa_with_dic():

    happy_qa = HappyQuestionAnswering()
    train_args = {'learning_rate': 0.01,  "num_train_epochs": 1}


    happy_qa.train("../data/qa/train-eval.csv" , args=train_args)

    eval_args = {}

    result_eval = happy_qa.eval("../data/qa/train-eval.csv", args=eval_args)
    assert type(result_eval.loss) == float

    test_args = {}

    result_test = happy_qa.test("../data/qa/test.csv", args=test_args)
    assert type(result_test[0].answer) == str

def test_tc_with_dataclass():

    happy_qa = HappyQuestionAnswering()
    train_args = QATrainArgs(learning_rate=0.01, num_train_epochs=1)

    happy_qa.train("../data/qa/train-eval.csv", args=train_args)

    eval_args = QAEvalArgs()

    result_eval = happy_qa.eval("../data/qa/train-eval.csv", args=eval_args)
    assert type(result_eval.loss) == float


    test_args = QATestArgs()

    result_test = happy_qa.test("../data/qa/test.csv", args=test_args)
    assert type(result_test[0].answer) == str
